#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include "PID.h"
#include "type_def.h"
#include "debug_rt.h"
#include "communication.h"
#include "computation_graph.h"
#include "timer.h"

/*
    Initially sets it so that all debug values are zero
*/
SensorRTFlags_t * initializeFlags(sensorRTF_flags_t * flags) {

    flags = calloc(1, sizeof(SensorRTFlags_t));
    flags -> imuflags = calloc(1, sizeof(IMUFlags_t));
    flags -> optflowflags = calloc(1, sizeof(OptFlowFlags_t));
    flags -> lidarflags = calloc(1, sizeof(lidarFlags_t));
    flags -> errorFlags = calloc(1, sizeof(SensorErrorFlags_t));
}

void freeFlags(SensorRTFlags_t * flags){
    free(flags -> imuflags);
    free(flags -> optflowflags); 
    free(flags -> lidarflags);
    free(flags -> errorFlags);
    free(flags);
}

/*
    Helper to return single bit of u32 data. This returns the "position"'th bit of the given u32,
    assuming it is Zero indexed.
*/
u32 read_bit(u32 data, int position){
    return (data >> position) & 1;
}

/*
Uses the payload from the configuration packet (it's in the flag parameter) to set what should be sent
Size of float = 32 bits

As of now many bits are empty and wont be processed
See RTSensorPacketDocumentation.md for which flags each bit of the config data corresponds to

It returns the size of the packet that will be sent, so the payload is correctly determined
*/
int process_configuration_packet(u32 configData, SensorRTFlags_t * flags)
{
    int size = 0;
    flags -> lidarflags -> quadHeight = read_bit(configData, 31); size += read_bit(configData, 31); 
    flags -> optflowflags ->  x_flow = read_bit(configData, 30); size += 2 * read_bit(configData, 30); 
    flags -> optflowflags ->  y_flow = read_bit(configData, 29); size += 2 * read_bit(configData, 29); 
    flags -> optflowflags ->  x_filter = read_bit(configData, 28); size += 2 * read_bit(configData, 28); 
    flags -> optflowflags ->  y_filter = read_bit(configData, 27); size += 2 * read_bit(configData, 27); 
    flags -> optflowflags ->  x_velocity = read_bit(configData, 26); size += 2 * read_bit(configData, 26); 
    flags -> optflowflags ->  y_velocity = read_bit(configData, 25); size += 2 * read_bit(configData, 25); 
    flags -> imuflags -> gyro_x = read_bit(configData, 24); size += read_bit(configData, 24); 
    flags -> imuflags -> gyro_y = read_bit(configData, 23); size += read_bit(configData, 23); 
    flags -> imuflags -> gyro_z = read_bit(configData, 22); size += read_bit(configData, 22); 
    flags -> imuflags -> acc_x = read_bit(configData, 21); size += read_bit(configData, 21); 
    flags -> imuflags -> acc_y = read_bit(configData, 20); size += read_bit(configData, 20); 
    flags -> imuflags -> acc_z = read_bit(configData, 19); size += read_bit(configData, 19); 
    flags -> imuflags -> mag_x = read_bit(configData, 18); size += read_bit(configData, 18); 
    flags -> imuflags -> mag_y = read_bit(configData, 17); size += read_bit(configData, 17); 
    flags -> imuflags -> mag_z = read_bit(configData, 16); size += read_bit(configData, 16); 
    flags -> errorflags -> lidar = read_bit(configData, 15); size += read_bit(configData, 15); 
    flags -> errorflags -> consec_lidar = read_bit(configData, 14); size += read_bit(configData, 14); 
    flags -> errorflags -> optFlow = read_bit(configData, 13); size += read_bit(configData, 13); 
    flags -> errorflags -> consec_optFlow = read_bit(configData, 12); size += read_bit(configData, 12); 
    flags -> errorflags -> imu = read_bit(configData, 11); size += read_bit(configData, 11); 
    flags -> errorflags -> consec_imu = read_bit(configData, 10); size += read_bit(configData, 10); 
    return size;
}

int send_RT_data(struct CommDriver *comm, raw_sensor_t * data, SensorRTFlags_t * flags, int size)
{
    char * payload = malloc(sizeof(char) * size);
    int currentBytePosition = 0;
    /*
    Since it's calloced (all zeros), a simple & operation should set this properly
    */
    if(flags -> lidarflags -> quadHeight){

        (*payload) &= ((data -> lidar -> distance_m) << (currentBytePosition * 8));
        currentBytePosition++; 
    } 

    if(flags -> optflowflags ->  x_flow) 
    {
        (*payload) &= ((data -> optical_flow -> flow_x_rad) << (currentBytePosition * 8));
        currentBytePosition += 2;
    }

    if(flags -> optflowflags ->  y_flow)
    {
        (*payload) &= ((data -> optical_flow -> flow_y_rad) << (currentBytePosition * 8));
        currentBytePosition += 2;
    }

    if(flags -> optflowflags ->  x_filter)
    {
        (*payload) &= ((data -> optical_flow -> xVelFilt) << (currentBytePosition * 8));
        currentBytePosition += 2;
    }

    if(flags -> optflowflags ->  y_filter)
    {
        (*payload) &= ((data -> optical_flow -> yVelFilt) << (currentBytePosition * 8));
        currentBytePosition += 2;
    }
    if(flags -> optflowflags ->  x_velocity)
    {
        (*payload) &= ((data -> optical_flow -> xVel) << (currentBytePosition * 8));
        currentBytePosition += 2;
    }

    if(flags -> optflowflags ->  y_velocity)
    {
        (*payload) &= ((data -> optical_flow -> yVel) << (currentBytePosition * 8));
        currentBytePosition += 2;
    }
    
    if(flags -> imuflags -> gyro_x)
    {
        (*payload) &= ((data -> gam -> gyro_xVel_p) << (currentBytePosition * 8));
        currentBytePosition++; 
    }

    if(flags -> imuflags -> gyro_y)
    {
        (*payload) &= ((data -> gam -> gyro_yVel_q) << (currentBytePosition * 8));
        currentBytePosition++; 
    }

    if(flags -> imuflags -> gyro_z)
    {
        (*payload) &= ((data -> gam -> gyro_zVel_r) << (currentBytePosition * 8));
        currentBytePosition++; 
    } 

    if(flags -> imuflags -> acc_x)
    {
        (*payload) &= ((data -> gam -> accel_x) << (currentBytePosition * 8));
        currentBytePosition++; 
    }

    if(flags -> imuflags -> acc_y)
    {
        (*payload) &= ((data -> gam -> accel_y) << (currentBytePosition * 8));
        currentBytePosition++; 
    }

    if(flags -> imuflags -> acc_z)
    {
        (*payload) &= ((data -> gam -> accel_z) << (currentBytePosition * 8));
        currentBytePosition++; 
    }

    if(flags -> imuflags -> mag_x)
    {
        (*payload) &= ((data -> gam -> mag_x) << (currentBytePosition * 8));
        currentBytePosition++; 
    }

    if(flags -> imuflags -> mag_y)
    {
        (*payload) &= ((data -> gam -> mag_y) << (currentBytePosition * 8));
        currentBytePosition++; 
    }

    if(flags -> imuflags -> mag_z)
    {
        (*payload) &= ((data -> gam -> mag_z) << (currentBytePosition * 8));
        currentBytePosition++; 
    }
    
    if(flags -> errorflags -> lidar)
    {
        (*payload) &= ((data -> lidar -> error -> errorCount) << (currentBytePosition * 8));
        currentBytePosition++; 
    }

    if(flags -> errorflags -> consec_lidar)
    {
        (*payload) &= ((data -> lidar -> consErrorCount -> errorCount) << (currentBytePosition * 8));
        currentBytePosition++; 
    }
    
    if(flags -> errorflags -> optFlow)
    {
        (*payload) &= ((data -> optical_flow -> error -> errorCount) << (currentBytePosition * 8));
        currentBytePosition++; 
    }

    if(flags -> errorflags -> consec_optFlow)
    {
        (*payload) &= ((data -> optical_flow -> consErrorCount -> errorCount) << (currentBytePosition * 8));
        currentBytePosition++; 
    }

    if(flags -> errorflags -> imu)
    {
        (*payload) &= ((data -> gam -> error -> errorCount) << (currentBytePosition * 8));
        currentBytePosition++;
    }
    
    if(flags -> errorflags -> consec_imu)
    {
        (*payload) &= ((data -> gam -> consErrorCount -> errorCount) << (currentBytePosition * 8));
        currentBytePosition++; 
    }

    int bytes_sent = send_data(comm->uart, SEND_RT_ID, 0, (u8*)payload, strlen(payload));
    free(payload);
    return bytes_sent; 
}
